@@ -24,28 +24,18 @@
 #include "bits.h"
 #include "dwg.h"
 
-enum RES_BUF_VALUE_TYPE
-{
-  VT_INVALID = 0,
-  VT_STRING = 1,
-  VT_POINT3D = 2,
-  VT_REAL = 3,
-  VT_INT16 = 4,
-  VT_INT32 = 5,
-  VT_INT8 = 6,
-  VT_BINARY = 7,
-  VT_HANDLE = 8,
-  VT_OBJECTID = 9,
-  VT_BOOL = 10,
-  VT_INT64 = 11, // BLL
-};
+#ifdef IN_DXF_H
+#error in_dxf.h must be included after decode.h because of FORMAT_BD
+#endif
 
-enum RES_BUF_VALUE_TYPE get_base_value_type (short gc);
+#define REFS_PER_REALLOC 128
 
-int dwg_decode (Bit_Chain *restrict dat, Dwg_Data *restrict dwg);
+EXPORT int dwg_decode (Bit_Chain *restrict dat, Dwg_Data *restrict dwg);
 int dwg_decode_unknown (Bit_Chain *restrict dat, Dwg_Object *restrict obj);
 Dwg_Object_Ref *dwg_find_objectref (const Dwg_Data *restrict dwg,
                                     const Dwg_Object *restrict obj);
+// reused with the importers
+void decode_BACKGROUND_type (const Dwg_Object *obj);
 
 /*------------------------------------------------------------------------------
  * Functions reused with decode_r2007
@@ -62,14 +52,21 @@ int dwg_decode_header_variables (Bit_Chain *dat, Bit_Chain *hdl_dat,
                                  Bit_Chain *str_dat, Dwg_Data *restrict dwg);
 int dwg_decode_add_object (Dwg_Data *restrict dwg, Bit_Chain *dat,
                            Bit_Chain *hdl_dat, long unsigned int address);
+int obj_handle_stream (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
+                       Bit_Chain *restrict hdl_dat);
 void bfr_read (void *restrict dst, BITCODE_RC *restrict *restrict src,
                size_t size);
+void decrypt_R2004_header (BITCODE_RC *restrict dest,
+                           const BITCODE_RC *restrict src, unsigned size);
 
 /* reused with free */
 void dwg_free_xdata_resbuf (Dwg_Resbuf *restrict rbuf);
 
 /* reused with encode */
 void dwg_resolve_objectrefs_silent (Dwg_Data *restrict dwg);
+uint32_t dwg_section_page_checksum (const uint32_t seed, Bit_Chain *restrict dat,
+                                    int32_t size);
+unsigned int section_max_decomp_size (const Dwg_Data *dwg, const Dwg_Section_Type id);
 
 /* reused with out_dxf */
 char *dwg_dim_blockname (Dwg_Data *restrict dwg,
@@ -77,10 +74,11 @@ char *dwg_dim_blockname (Dwg_Data *restrict dwg,
 
 int dwg_validate_INSERT (Dwg_Object *restrict obj);
 int dwg_validate_POLYLINE (Dwg_Object *restrict obj);
+/* reused with many */
+int dwg_fixup_BLOCKS_entities (Dwg_Data *restrict dwg);
+void dxf_3dsolid_revisionguid (Dwg_Entity_3DSOLID *_obj);
 
 /* from decode_r2007.c */
-int obj_handle_stream (Bit_Chain *dat, Dwg_Object *restrict obj,
-                       Bit_Chain *hdl_dat);
 int obj_string_stream (Bit_Chain *dat, Dwg_Object *restrict obj,
                        Bit_Chain *str_dat);
 unsigned long obj_stream_position (Bit_Chain *restrict dat,
@@ -93,4 +91,11 @@ int read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
 void section_string_stream (Bit_Chain *restrict dat, BITCODE_RL bitsize,
                             Bit_Chain *restrict str);
 
+/* from dwg.c */
+// from dat.fh
+EXPORT int dat_read_size (Bit_Chain *restrict dat);
+EXPORT int dat_read_file (Bit_Chain *restrict dat, FILE *restrict fp,
+                          const char *restrict filename);
+EXPORT int dat_read_stream (Bit_Chain *restrict dat, FILE *restrict fp);
+
 #endif
