@@ -16,19 +16,20 @@
  * modified by Reini Urban
  */
 
+#define IS_DECODER
+
 #include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <stdint.h>
 #include <sys/types.h>
 #include <stdbool.h>
 #include <assert.h>
 #include "common.h"
 #include "bits.h"
-#include "dec_macros.h"
 #include "decode.h"
-
-#define MIN(X, Y) ((X) < (Y) ? (X) : (Y))
+#include "dynapi.h"
 
 /* The logging level for the read (decode) path.  */
 static unsigned int loglevel;
@@ -37,6 +38,7 @@ static unsigned int cur_ver = 0;
 
 #define DWG_LOGLEVEL loglevel
 #include "logging.h"
+#include "dec_macros.h"
 
 // only for temp. debugging, to abort on obviously wrong sizes.
 // should be a bit larger then the filesize.
@@ -609,6 +611,7 @@ decode_rs (const BITCODE_RC *src, int block_count, int data_size,
   return dst_base;
 }
 
+ATTRIBUTE_MALLOC
 static BITCODE_RC *
 read_system_page (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
                   int64_t repeat_count)
@@ -629,9 +632,26 @@ read_system_page (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
   pesize = ((size_comp + 7) & ~7) * repeat_count;
   // Divide pre encoded size by RS k-value (239)
   block_count = (pesize + 238) / 239;
+  if (!block_count)
+    {
+      LOG_ERROR ("Empty r2007 system page block_count. size_comp: %" PRId64
+                 ", repeat_count: %" PRId64,
+                 size_comp, repeat_count);
+      return NULL;
+    }
+
   // Multiply with codeword size (255) and round to a multiple of 8
   page_size = (block_count * 255 + 7) & ~7;
 
+  if (!((uint64_t)size_comp < dat->size &&
+        (uint64_t)size_uncomp < dat->size &&
+        (uint64_t)repeat_count < DBG_MAX_COUNT &&
+        (uint64_t)page_size < DBG_MAX_COUNT))
+    {
+      LOG_ERROR ("Invalid r2007 system page size_comp: %" PRId64
+                 ", size_uncomp: %" PRId64, size_comp, size_uncomp);
+      return NULL;
+    }
   assert ((uint64_t)size_comp < dat->size);
   assert ((uint64_t)size_uncomp < dat->size);
   assert ((uint64_t)repeat_count < DBG_MAX_COUNT);
@@ -644,6 +664,7 @@ read_system_page (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       return NULL;
     }
   data = (BITCODE_RC *)calloc (size_uncomp + page_size, 1);
+  LOG_HANDLE ("Alloc system page of size %" PRId64 "\n", size_uncomp + page_size)
   if (!data)
     {
       LOG_ERROR ("Out of memory")
@@ -664,8 +685,8 @@ read_system_page (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
     error = decompress_r2007 (data, size_uncomp, pedata, MIN (pedata_size, size_comp));
   else
     memcpy (data, pedata, size_uncomp);
-  free (pedata);
 
+  free (pedata);
   if (error >= DWG_ERR_CRITICAL)
     {
       free (data);
@@ -702,7 +723,10 @@ read_data_page (Bit_Chain *restrict dat, BITCODE_RC *restrict decomp,
   bit_read_fixed (dat, rsdata, page_size);
   pedata = decode_rs (rsdata, block_count, 0xFB, page_size);
   if (!pedata)
-    return DWG_ERR_OUTOFMEM;
+    {
+      free (rsdata);
+      return DWG_ERR_OUTOFMEM;
+    }
 
   if (size_comp < size_uncomp)
     error = decompress_r2007 (decomp, size_uncomp, pedata,
@@ -728,10 +752,21 @@ read_data_section (Bit_Chain *sec_dat, Bit_Chain *dat,
   int error = 0, i;
 
   section = get_section (sections_map, sec_type);
+  sec_dat->chain = NULL;
   if (section == NULL)
     {
-      LOG_ERROR ("Failed to find section %d", (int)sec_type)
-      return DWG_ERR_SECTIONNOTFOUND;
+      if (sec_type < SECTION_REVHISTORY
+          && sec_type != SECTION_TEMPLATE
+          && sec_type != SECTION_OBJFREESPACE)
+        {
+          LOG_WARN ("Failed to find section_info[%u]", (int)sec_type)
+          return DWG_ERR_SECTIONNOTFOUND;
+        }
+      else
+        {
+          LOG_TRACE ("Found no section_info[%u]\n", (int)sec_type)
+          return DWG_ERR_VALUEOUTOFBOUNDS;
+        }
     }
 
   max_decomp_size = section->data_size;
@@ -740,17 +775,19 @@ read_data_section (Bit_Chain *sec_dat, Bit_Chain *dat,
       LOG_ERROR ("Invalid max decompression size %" PRIu64, max_decomp_size);
       return DWG_ERR_INVALIDDWG;
     }
-  decomp = calloc (max_decomp_size, 1);
+  decomp = (BITCODE_RC *)calloc (max_decomp_size, 1);
   if (decomp == NULL)
     {
       LOG_ERROR ("Out of memory")
       return DWG_ERR_OUTOFMEM;
     }
+  LOG_HANDLE ("Alloc data section of size %" PRIu64 "\n", max_decomp_size)
 
   sec_dat->bit = 0;
   sec_dat->byte = 0;
   sec_dat->size = max_decomp_size;
   sec_dat->version = dat->version;
+  sec_dat->from_version = dat->from_version;
 
   for (i = 0; i < (int)section->num_pages; i++)
     {
@@ -869,6 +906,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       if (!section)
         {
           LOG_ERROR ("Out of memory");
+          free (data);
           sections_destroy (sections); // the root
           return NULL;
         }
@@ -919,7 +957,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       if (ptr >= ptr_end)
         break;
 
-      // Section Name
+      // Section Name (wchar)
       section->name = bfr_read_string (&ptr, section->name_length);
 #ifdef HAVE_NATIVE_WCHAR2
       LOG_TRACE ("  name:          " FORMAT_TU "\n\n",
@@ -929,7 +967,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       LOG_TEXT_UNICODE (TRACE, section->name)
       LOG_TRACE ("\n\n")
 #endif
-      section->type = dwg_section_type (section->name);
+      section->type = dwg_section_wtype (section->name);
 
       if (section->num_pages <= 0)
         continue;
@@ -944,6 +982,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
       if (!section->pages)
         {
           LOG_ERROR ("Out of memory");
+          free (data);
           if (sections)
             sections_destroy (sections); // the root
           else
@@ -958,6 +997,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
           if (!section->pages[i])
             {
               LOG_ERROR ("Out of memory");
+              free (data);
               if (sections)
                 sections_destroy (sections); // the root
               else
@@ -968,6 +1008,7 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
           if (ptr + 56 >= ptr_end)
             {
               LOG_ERROR ("Section[%d]->pages[%d] overflow", j, i);
+              free (section->pages[i]);
               section->num_pages = i; // skip this last section
               break;
             }
@@ -987,7 +1028,9 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
               || section->pages[i]->uncomp_size >= DBG_MAX_SIZE
               || section->pages[i]->comp_size >= DBG_MAX_SIZE)
             {
-              LOG_ERROR ("Invalid section->pages[%d].*size", i);
+              LOG_ERROR ("Invalid section->pages[%d] size", i);
+              free (data);
+              free (section->pages[i]);
               section->num_pages = i; // skip this last section
               return sections;
             }
@@ -998,7 +1041,6 @@ read_sections_map (Bit_Chain *dat, int64_t size_comp, int64_t size_uncomp,
     }
 
   free (data);
-
   return sections;
 }
 
@@ -1227,44 +1269,20 @@ obj_stream_position (Bit_Chain *restrict dat, Bit_Chain *restrict hdl_dat,
                      Bit_Chain *restrict str_dat)
 {
   unsigned long p1 = bit_position (dat);
+  /* all 3 now relative to obj */
+  unsigned long p2 = bit_position (hdl_dat);
   SINCE (R_2007)
-  {
-    /* all 3 now relative to obj */
-    unsigned long p2 = bit_position (hdl_dat);
-    unsigned long p3 = bit_position (str_dat);
-
-    if (p2 > p1)
-      return p3 > p2 ? p3 : p2;
-    else
-      return p3 > p1 ? p3 : p1;
-  }
-  else { return p1; }
-}
-
-int
-obj_handle_stream (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
-                   Bit_Chain *restrict hdl_dat)
-{
-  long unsigned int bit8 = obj->bitsize / 8;
-  long unsigned int pos = dat->byte;
-  // The handle stream offset, i.e. end of the object, right after
-  // the has_strings bit.
-  obj->hdlpos = obj->bitsize; // relative to dat
-  // restrict it to 0-end
-  hdl_dat->byte = bit8;
-  hdl_dat->bit = obj->bitsize % 8;
-  // bit_reset_chain (hdl_dat); //but keep the same start
-  if (!obj->handlestream_size)
+    { // but only since 2007 there is a seperate string stream
+      unsigned long p3 = bit_position (str_dat);
+      if (p2 > p1)
+        return p3 > p2 ? p3 : p2;
+      else
+        return p3 > p1 ? p3 : p1;
+    }
+  else
     {
-      obj->handlestream_size = (obj->size * 8) - obj->bitsize;
-      LOG_TRACE (" Hdlsize: %lu,", obj->handlestream_size);
+      return (p2 > p1) ? p2 : p1;
     }
-  hdl_dat->size = obj->size;
-  pos = (pos * 8) + obj->bitsize + obj->handlestream_size;
-  LOG_HANDLE (" hdl_dat: @%lu.%u - @%lu.%lu (%lu)", bit8, hdl_dat->bit,
-              pos / 8, pos % 8, hdl_dat->size);
-  LOG_TRACE ("\n")
-  return 0;
 }
 
 int
@@ -1275,6 +1293,7 @@ obj_string_stream (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
   BITCODE_RL data_size = 0;            // in byte
   BITCODE_RL old_size;                 // in byte
   BITCODE_RL old_byte;
+  assert (dat != str);
   old_size = str->size;
   old_byte = str->byte;
 
@@ -1288,7 +1307,12 @@ obj_string_stream (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
 
   if (str->byte >= old_size - old_byte)
     {
-      LOG_WARN ("obj_string_stream overflow");
+      LOG_ERROR ("obj_string_stream overflow, bitsize " FORMAT_RL " => " FORMAT_RL,
+                 obj->bitsize, obj->size * 8);
+      str->byte = old_byte;
+      str->size = old_size;
+      obj->has_strings = 0;
+      obj->bitsize = obj->size * 8;
       return DWG_ERR_VALUEOUTOFBOUNDS;
     }
   LOG_HANDLE (" obj string stream +%u: @%lu.%u (%lu)", start, str->byte,
@@ -1297,7 +1321,10 @@ obj_string_stream (Bit_Chain *restrict dat, Dwg_Object *restrict obj,
   LOG_TRACE (" has_strings: %d\n", (int)obj->has_strings);
   if (!obj->has_strings)
     {
-      str->size = 0;
+      // FIXME wrong bit
+      if (obj->fixedtype == DWG_TYPE_SCALE)
+        obj->has_strings = 1;
+      // str->size = 0;
       // bit_reset_chain (str);
       return 0;
     }
@@ -1417,13 +1444,13 @@ read_2007_section_classes (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
       size = bit_read_RL (&sec_dat); // size of class data area
       LOG_TRACE ("size: " FORMAT_RL " [RL]\n", size)
       /*
-      if (dat->version >= R_2010 && dwg->header.maint_version > 3)
+      if (dat->from_version >= R_2010 && dwg->header.maint_version > 3)
         {
           BITCODE_RL hsize = bit_read_RL(&sec_dat);
           LOG_TRACE("hsize: " FORMAT_RL " [RL]\n", hsize)
         }
       */
-      if (dat->version >= R_2007)
+      if (dat->from_version >= R_2007)
         {
           bitsize = bit_read_RL (&sec_dat);
           LOG_TRACE ("bitsize: " FORMAT_RL " [RL]\n", bitsize)
@@ -1443,6 +1470,8 @@ read_2007_section_classes (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
         {
           LOG_ERROR ("Invalid max class number %d", max_num)
           dwg->num_classes = 0;
+          if (sec_dat.chain)
+            free (sec_dat.chain);
           return DWG_ERR_VALUEOUTOFBOUNDS;
         }
       assert (max_num >= 500);
@@ -1539,14 +1568,14 @@ read_2007_section_header (Bit_Chain *restrict dat, Bit_Chain *restrict hdl_dat,
       *hdl_dat = sec_dat;
       // unused: later versions re-use the 2004 section format
       /*
-      if (dat->version >= R_2010 && dwg->header.maint_version > 3)
+      if (dat->from_version >= R_2010 && dwg->header.maint_version > 3)
         {
           dwg->header_vars.bitsize_hi = bit_read_RL(&sec_dat);
           LOG_TRACE("bitsize_hi: " FORMAT_RL " [RL]\n",
       dwg->header_vars.bitsize_hi) endbits += 32;
         }
       */
-      if (dat->version == R_2007) // always true so far
+      if (dat->from_version == R_2007) // always true so far
         {
           dwg->header_vars.bitsize = bit_read_RL (&sec_dat);
           LOG_TRACE ("bitsize: " FORMAT_RL " [RL]\n",
@@ -1683,41 +1712,519 @@ read_2007_section_handles (Bit_Chain *dat, Bit_Chain *hdl,
   return error;
 }
 
+/* VBAProject Section
+ */
+static int
+read_2007_section_vbaproject (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                              r2007_section *restrict sections_map,
+                              r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  //Bit_Chain *str_dat;
+  Dwg_VBAProject *_obj = &dwg->vbaproject;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  //BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // not compressed, page size: 0x80
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_VBAPROJECT);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "VBAProject");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error;
+    }
+
+  LOG_TRACE ("\nVBAProject (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  DEBUG_HERE
+  _obj->size = dat->size;
+  _obj->unknown_bits = bit_read_TF (dat, _obj->size);
+  LOG_TRACE_TF (_obj->unknown_bits, _obj->size)
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
 static int
 read_2007_section_summary (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
                            r2007_section *restrict sections_map,
                            r2007_page *restrict pages_map)
 {
-  static Bit_Chain old_dat, obj_dat = { 0 };
+  static Bit_Chain old_dat, sec_dat = { 0 };
   Bit_Chain *str_dat;
-  struct Dwg_SummaryInfo *_obj = &dwg->summaryinfo;
+  Dwg_SummaryInfo *_obj = &dwg->summaryinfo;
   Dwg_Object *obj = NULL;
   int error;
   BITCODE_RL rcount1 = 0, rcount2 = 0;
 
   old_dat = *dat;
-  error = read_data_section (&obj_dat, dat, sections_map, pages_map,
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
                              SECTION_SUMMARYINFO);
-  if (error >= DWG_ERR_CRITICAL || !obj_dat.chain)
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
     {
       LOG_ERROR ("Failed to read SummaryInfo section");
-      if (obj_dat.chain)
-        free (obj_dat.chain);
+      if (sec_dat.chain)
+        free (sec_dat.chain);
       return error;
     }
 
   if (dwg->header.summaryinfo_address != (BITCODE_RL)dat->byte)
     LOG_WARN ("summaryinfo_address mismatch: " FORMAT_RL " != %lu",
               dwg->header.summaryinfo_address, dat->byte);
-  LOG_TRACE ("\nSummaryInfo\n-------------------\n")
-  str_dat = dat = &obj_dat; // restrict in size
+  LOG_TRACE ("\nSummaryInfo (%lu)\n-------------------\n", sec_dat.size)
+  str_dat = dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
 
   // clang-format off
   #include "summaryinfo.spec"
   // clang-format on
 
-  if (obj_dat.chain)
-    free (obj_dat.chain);
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* AppInfo Section
+ */
+static int
+read_2007_section_appinfo (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                           r2007_section *restrict sections_map,
+                           r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  Bit_Chain *str_dat;
+  Dwg_AppInfo *_obj = &dwg->appinfo;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // not compressed, page size: 0x80
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_APPINFO);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "AppInfo");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error;
+    }
+
+  LOG_TRACE ("\nAppInfo (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  str_dat = dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "appinfo.spec"
+  // clang-format on
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* AuxHeader Section
+ */
+static int
+read_2007_section_auxheader (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                             r2007_section *restrict sections_map,
+                             r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  //Bit_Chain *str_dat;
+  Dwg_AuxHeader *_obj = &dwg->auxheader;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  BITCODE_RL vcount = 0, rcount1 = 0, rcount2 = 0;
+
+  // type: 2, compressed, page size: 0x7400
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_AUXHEADER);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "AuxHeader");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error;
+    }
+
+  LOG_TRACE ("\nAuxHeader (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "auxheader.spec"
+  // clang-format on
+  
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* Unknown AppInfoHistory Section
+ */
+static int
+read_2007_section_appinfohistory (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                                  r2007_section *restrict sections_map,
+                                  r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  //Bit_Chain *str_dat;
+  Dwg_AppInfoHistory *_obj = &dwg->appinfohistory;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  //BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // compressed, page size: 0x580
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_APPINFOHISTORY);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "AppInfoHistory");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error;
+    }
+
+  LOG_TRACE ("\nAppInfoHistory (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  DEBUG_HERE
+  _obj->size = dat->size;
+  _obj->unknown_bits = bit_read_TF (dat, _obj->size);
+  LOG_TRACE_TF (_obj->unknown_bits, _obj->size)
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* RevHistory Section
+ */
+static int
+read_2007_section_revhistory (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                              r2007_section *restrict sections_map,
+                              r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  //Bit_Chain *str_dat;
+  Dwg_RevHistory *_obj = &dwg->revhistory;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // compressed, page size: 0x7400
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_REVHISTORY);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "RevHistory");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error;
+    }
+
+  LOG_TRACE ("\nRevHistory (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "revhistory.spec"
+  // clang-format on
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* ObjFreeSpace Section
+ */
+static int
+read_2007_section_objfreespace (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                                r2007_section *restrict sections_map,
+                                r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  //Bit_Chain *str_dat;
+  Dwg_ObjFreeSpace *_obj = &dwg->objfreespace;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // compressed, page size: 0x7400
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_OBJFREESPACE);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "ObjFreeSpace");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error;
+    }
+
+  LOG_TRACE ("\nObjFreeSpace (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "objfreespace.spec"
+  // clang-format on
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* Template Section. Optional r13-r15, mandatory r18+.
+   Contains the MEASUREMENT variable (0 = English, 1 = Metric).
+ */
+static int
+read_2007_section_template (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                            r2007_section *restrict sections_map,
+                            r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  //Bit_Chain *str_dat;
+  Dwg_Template *_obj = &dwg->Template;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // compressed
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_TEMPLATE);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_ERROR ("%s section not found\n", "Template");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return error | DWG_ERR_SECTIONNOTFOUND;
+    }
+
+  LOG_TRACE ("\nTemplate (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "template.spec"
+  // clang-format on
+
+  dwg->header_vars.MEASUREMENT = _obj->MEASUREMENT;
+  LOG_TRACE ("HEADER.MEASUREMENT: " FORMAT_BS " (0 English/1 Metric)\n",
+             dwg->header_vars.MEASUREMENT)
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* r21 FileDepList Section
+ */
+static int
+read_2007_section_filedeplist (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                               r2007_section *restrict sections_map,
+                               r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  int error;
+  Bit_Chain *str_dat;
+  Dwg_FileDepList *_obj = &dwg->filedeplist;
+  Dwg_Object *obj = NULL;
+  BITCODE_BL vcount;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // not compressed, page size: 0x80. 0xc or 0xd
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_FILEDEPLIST);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "FileDepList");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return 0;
+    }
+
+  LOG_TRACE ("FileDepList (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  str_dat = dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "filedeplist.spec"
+  // clang-format on
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+/* r21 Security Section, if saved with password
+ */
+static int
+read_2007_section_security (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                           r2007_section *restrict sections_map,
+                           r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  int error;
+  Bit_Chain *str_dat;
+  Dwg_Security *_obj = &dwg->security;
+  Dwg_Object *obj = NULL;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // compressed, page size: 0x7400
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_SECURITY);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "Security");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return 0;
+    }
+
+  LOG_TRACE ("Security (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  str_dat = dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "security.spec"
+  // clang-format on
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+
+#if 0
+/* Signature Section, not written nor documented by Teigha
+ */
+static int
+read_2007_section_signature (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                           r2007_section *restrict sections_map,
+                           r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  int error;
+  Bit_Chain *str_dat;
+  struct Dwg_Signature *_obj = &dwg->signature;
+  Dwg_Object *obj = NULL;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // compressed, page size: 0x7400
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map,
+                             SECTION_SIGNATURE);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", "Signature");
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return 0;
+    }
+
+  LOG_TRACE ("Signature (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  str_dat = dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  // clang-format off
+  #include "signature.spec"
+  // clang-format on
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
+  *dat = old_dat; // unrestrict
+  return error;
+}
+#endif
+
+static int
+acds_private (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
+{
+  Bit_Chain *str_dat = dat;
+  Dwg_AcDs *_obj = &dwg->acds;
+  Dwg_Object *obj = NULL;
+  int error = 0;
+  BITCODE_BL rcount1 = 0, rcount2 = 0;
+  BITCODE_BL rcount3 = 0, rcount4, vcount;
+
+  // clang-format off
+  #include "acds.spec"
+  // clang-format on
+
+  return error;
+}
+
+/* r2013+ datastorage Section, if saved with binary ACIS SAB data
+ */
+static int
+read_2007_section_acds (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
+                        r2007_section *restrict sections_map,
+                        r2007_page *restrict pages_map)
+{
+  Bit_Chain old_dat, sec_dat = { 0 };
+  int error;
+  Bit_Chain *str_dat;
+  const char *secname = "AcDsPrototype_1b";
+  Dwg_AcDs *_obj = &dwg->acds;
+  Dwg_Object *obj = NULL;
+  BITCODE_RL rcount1 = 0, rcount2 = 0;
+
+  // compressed, pagesize 0x7400, type 13
+  error = read_data_section (&sec_dat, dat, sections_map, pages_map, SECTION_ACDS);
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
+    {
+      LOG_INFO ("%s section not found\n", secname);
+      if (sec_dat.chain)
+        free (sec_dat.chain);
+      return 0;
+    }
+
+  LOG_TRACE ("AcDs datastorage (%lu)\n-------------------\n", sec_dat.size)
+  old_dat = *dat;
+  str_dat = dat = &sec_dat; // restrict in size
+  bit_chain_set_version (&old_dat, dat);
+
+  error |= acds_private (dat, dwg);
+  error &= ~DWG_ERR_SECTIONNOTFOUND;
+
+  LOG_TRACE ("\n")
+  if (sec_dat.chain)
+    free (sec_dat.chain);
   *dat = old_dat; // unrestrict
   return error;
 }
@@ -1734,7 +2241,7 @@ read_2007_section_preview (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
 
   error = read_data_section (&sec_dat, dat, sections_map, pages_map,
                              SECTION_PREVIEW);
-  if (error >= DWG_ERR_CRITICAL)
+  if (error >= DWG_ERR_CRITICAL || !sec_dat.chain)
     {
       LOG_ERROR ("Failed to read uncompressed %s section", "Preview");
       if (sec_dat.chain)
@@ -1745,8 +2252,8 @@ read_2007_section_preview (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
   if (dwg->header.thumbnail_address != (BITCODE_RL)dat->byte)
     LOG_WARN ("thumbnail_address mismatch: " FORMAT_RL " != %lu",
               dwg->header.thumbnail_address, dat->byte);
-  LOG_TRACE ("\nPreview\n-------------------\n")
-  if (!sec_dat.chain || sec_dat.size < 16)
+  LOG_TRACE ("\nPreview (%lu)\n-------------------\n", sec_dat.size)
+  if (!sec_dat.chain || sec_dat.size < 32)
     {
       LOG_WARN ("Empty thumbnail");
       if (sec_dat.chain)
@@ -1761,6 +2268,8 @@ read_2007_section_preview (Bit_Chain *restrict dat, Dwg_Data *restrict dwg,
       return error |= DWG_ERR_WRONGCRC;
     }
 
+  assert (sec_dat.size >= 32);
+  assert (sec_dat.chain);
   dwg->thumbnail.size = sec_dat.size - 32; // 2x sentinel
   dwg->thumbnail.chain = sec_dat.chain;
   dwg->thumbnail.byte = 16; // sentinel
@@ -1789,7 +2298,7 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
                       Dwg_Data *restrict dwg)
 {
   r2007_file_header file_header;
-  r2007_page *restrict pages_map, *restrict page;
+  r2007_page *restrict pages_map = NULL, *restrict page;
   r2007_section *restrict sections_map = NULL;
   int error;
 #ifdef USE_TRACING
@@ -1810,7 +2319,16 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
   // Pages Map
   dat->byte += 0x28; // overread check data
   dat->byte += file_header.pages_map_offset;
-
+  if ((unsigned long)file_header.pages_map_size_comp
+      > dat->size - dat->byte)
+    {
+      LOG_ERROR ("%s Invalid pages_map_size_comp %lu > %lu bytes left",
+                 __FUNCTION__,
+                 (unsigned long)file_header.pages_map_size_comp,
+                 dat->size - dat->byte)
+      error |= DWG_ERR_VALUEOUTOFBOUNDS;
+      goto error;
+    }
   pages_map = read_pages_map (dat, file_header.pages_map_size_comp,
                               file_header.pages_map_size_uncomp,
                               file_header.pages_map_correction);
@@ -1828,7 +2346,7 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
     }
   dat->byte = page->offset;
   if ((unsigned long)file_header.sections_map_size_comp
-      > dat->byte - dat->size)
+      > dat->size - dat->byte)
     {
       LOG_ERROR ("%s Invalid comp_data_size %lu > %lu bytes left",
                  __FUNCTION__,
@@ -1850,16 +2368,20 @@ read_r2007_meta_data (Bit_Chain *dat, Bit_Chain *hdl_dat,
   error |= read_2007_section_classes (dat, dwg, sections_map, pages_map);
   error |= read_2007_section_handles (dat, hdl_dat, dwg, sections_map,
                                       pages_map);
+  error |= read_2007_section_auxheader (dat, dwg, sections_map, pages_map);
   if (dwg->header.thumbnail_address)
     error |= read_2007_section_preview (dat, dwg, sections_map, pages_map);
-  // if (dwg->header.vbaproj_address)
-  //  error |= read_2007_section_vbaproject (dat, dwg, sections_map,
-  //  pages_map);
-  // error |= read_2007_section_appinfo (dat, dwg, sections_map, pages_map);
-  // error |= read_2007_section_filedeplist (dat, dwg, sections_map,
-  // pages_map); error |= read_2007_section_security (dat, dwg, sections_map,
-  // pages_map); error |= read_2007_section_revhistory (dat, dwg, sections_map,
-  // pages_map);
+  error |= read_2007_section_appinfo (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_appinfohistory (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_filedeplist (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_security (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_revhistory (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_objfreespace (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_template (dat, dwg, sections_map, pages_map);
+  if (dwg->header.vbaproj_address)
+    error |= read_2007_section_vbaproject (dat, dwg, sections_map, pages_map);
+  //error |= read_2007_section_signature (dat, dwg, sections_map, pages_map);
+  error |= read_2007_section_acds (dat, dwg, sections_map, pages_map);
   // read_2007_blocks (dat, hdl_dat, dwg, sections_map, pages_map);
 
 error:
